#! /usr/bin/env python


from ..common.device_switch import Device_Switch
from .device_zwave_base import Device_ZWave_Base

paddle_events = {"DON", "DOF", "DIM", "BRT", "DFON", "DFOF"}


class Device_ZWave_GenericRspCtl(Device_Switch, Device_ZWave_Base):
    def __init__(self, container, device_info):
        Device_Switch.__init__(self, container, device_info.name, device_info.address)
        Device_ZWave_Base.__init__(self, device_info)

        self.add_property("paddle_action")

        value = device_info.get_property("ST", "value")
        if value:
            try:
                if int(value) > 0:
                    self.set_property("onoff", "on")
                else:
                    self.set_property("onoff", "off")
            except:
                pass

    def process_websocket_event(self, event):
        if event.control == "ST":
            if int(event.action) > 0:
                self.set_property("onoff", "on")
            else:
                self.set_property("onoff", "off")

        elif event.control in paddle_events:  # need to add other events
            self.set_property("paddle_action", event.control, True)

    def turn_on(self):
        path = "nodes/" + self.address + "/cmd/DON"
        return self.send_request(path)

    def turn_off(self):
        path = "nodes/" + self.address + "/cmd/DOF"
        return self.send_request(path)



#! /usr/bin/env python

""" 

Base Program

<eventInfo>
    <id>1C</id>
    <on />
    <nr />
    <r>190520 08:28:16</r>
    <f>190520 08:28:16</f>
    <s>31</s>
</eventInfo>

"""
from ..item_base import Item_Base


class Program_Base(Item_Base):
    def __init__(self, container, program_info):
        Item_Base.__init__(self, container, program_info.name)

        self.id = program_info.id

        self.add_property("state", "idle")  # program run state
        self.add_property("last_run_time", program_info.last_run_time)
        self.add_property("last_finish_time", program_info.last_finish_time)

    def __str__(self):
        return "Program: {}; id {}; run {}, finish {}".format(
            self.name,
            self.id,
            self.properties["last_run_time"],
            self.properties["last_finish_time"],
        )

    def process_websocket_event(self, state_code, last_run_time, last_finish_time):
        state = "unknown"
        if state_code & 1 == 1:
            state = "idle"
        elif state_code & 2 == 2:
            state = "then"
        elif state_code & 3 == 3:
            state = "else"

        self.set_property("state", state)
        self.set_property("last_run_time", last_run_time)
        self.set_property("last_finish_time", last_finish_time)

    def command(self, command):
        path = "programs/" + self.id + "/" + command
        return self.send_request(path)

    def run(self):
        self.command("run")

    def stop(self):
        self.command("stop")

    def run_then(self):
        self.command("runThen")

    def run_else(self):
        self.command("runElse")

    def enable(self):
        self.command("enable")

    def disable(self):
        self.command("disable")

    def get_identifier(self):
        return self.id




#! /usr/bin/env python

class Variable_Name(object):
    def __init__(self, node):

        self.valid = False

        try:

            self.id = node.attrib["id"]
            self.name = node.attrib["name"]
            self.value = int(node.find("val").text)
            self.valid = True

        except:
            pass

    def __repr__(self):
        return "Variable: ID {} Name {}, Value {}".format(
            self.id, self.name, self.value
        )




#! /usr/bin/env python

import time
import traceback
import threading
from datetime import datetime
from datetime import timedelta
import asyncio 

from .items.devices.device_container import Device_Container
from .items.scenes.scene_container import Scene_Container
from .items.variables.variable_container import Variable_Container
from .items.programs.program_container import Program_Container
from .items.controller.controller_container import Controller_Container

from .network.discover import Discover
from .network.async_session import Async_Session

from .support.repeating_timer import Repeating_Timer

import logging

logger = logging.getLogger(__name__)

# event categores controller, device, scene, variable, program


class Controller(object):
    def __init__(
        self,
        address=None,
        port=80,
        username="admin",
        password="admin",
        use_https=False,
        event_handler=None,
    ):
        if address == None:
            discover = Discover()
            controller_list = discover.start()
            if len(controller_list) > 0:
                address = controller_list[0]

        if address is None:
            logger.error("No controller address found")
            quit()

        self.address = address
        self.port = port
        self.username = username
        self.password = password
        self.use_https = use_https

        self.event_handler = event_handler

        self.controller_container = Controller_Container(self)
        self.device_container = Device_Container(self)
        self.scene_container = Scene_Container(self)
        self.variable_container = Variable_Container(self)
        self.program_container = Program_Container(self)

        self.last_heartbeat = datetime.now()
        self.heartbeat_interval = 30  # set below by data from controller, needed here for watchdog if no initial connection

        self.controller_container.start()

        self.connect()
        self.start()

        self.watch_dog_timer = Repeating_Timer(
            30
        )  # left over from websocket client - may remove as it does nothing
        self.watch_dog_timer.add_callback(self.watch_dog_check)

    def connect(self):
        def start():
            try:
                asyncio.set_event_loop(self.event_loop)
                self.event_loop.run_forever()
                logger.warning ('Event loop stopped')
                #self.session.close()
            except:
                logger.error ('Error in event loop')

        self.event_loop = asyncio.new_event_loop()

        self.session = Async_Session(
            self,
            self.address,
            self.port,
            self.username,
            self.password,
            False,
            loop=self.event_loop,
        )

        logger.warning("Starting Session thread")
        self._ws_thread = threading.Thread(target=start, args=())

        self._ws_thread.daemon = True
        self._ws_thread.start()

    def start(self):
        logger.info ('Starting controller')
        self.process_controller_event("status", "init")

        if self.get_controller_items() is True:
            self.process_controller_event("status", "ready")
            self.connect_websocket()
        else:
            self.process_controller_event("status", "error")
            self.retry_start(10)

    def retry_start(self, delay_seconds):
        def restart():
            self.start()

        self.restart_timer = threading.Timer(delay_seconds, restart)
        self.restart_timer.start()

    def get_controller_items(self):
        success = True

        if self.device_container.items_retrieved is False:
            if self.device_container.start() is False:
                success = False
                logger.info ('Get device items success {}'.format(success))

        if self.scene_container.items_retrieved is False:
            if self.scene_container.start() is False:
                success = False
                logger.info ('Get scene items success {}'.format(success))

        if self.variable_container.items_retrieved is False:
            if self.variable_container.start() is False:
                success = False
                logger.info ('Get variable items success {}'.format(success))

        if self.program_container.items_retrieved is False:
            if self.program_container.start() is False:
                success = False
                logger.info ('Get program items success {}'.format(success))

        succ, _ = self.controller_container.get_controller_time()
        if succ is False:
            success = False

        logger.info ('Retrieved controller items, success = {}'.format(success))

        return success

    def connect_websocket(self):
        logger.info ('Starting Websocket')
        self.session.start_websocket()
        logger.info ('Started Websocket')

    def container_event(self, container, item, event, *args):
        logger.debug(
            "Container Event {} from .{}: {} {}".format(
                item.name, container.container_type, item, args
            )
        )
        self.publish_container_event(container, item, event, *args)

        if (
            container.container_type == "Device"
        ):  # propagate to scene container to see if we should update scene states
            self.scene_container.device_event(item)

    def publish_container_event(self, container, item, event, *args):
        try:
            self.event_handler(container, item, event, *args)
        except Exception as ex:
            logger.error("Event handler Error {}".format(ex))
            traceback.print_exc()

    def send_request(self, path, timeout=None):
        success, response = self.session.request(path, timeout)
        return success, response

    def http_connected(self, connected):  # True HTTP connected, False, no connection
        if connected:
            self.process_controller_event("http", "connected")
        else:
            self.process_controller_event("http", "disconnected")

    def websocket_connected(
        self, connected
    ):  # True websocket connected, False, no connection
        if connected:
            self.process_controller_event("websocket", "connected")
        else:
            self.process_controller_event("websocket", "disconnected")

    def websocket_event(self, event):  # process websocket event
        logger.info("Websocket Event {}".format(event))

        try:
            if event.address is not None:  # event from .a device/node
                self.device_container.websocket_event(event)

            if event.control == "_0":  # heartbeat
                self.process_heartbeat(event)

            elif event.control == "_1":  # trigger
                if event.action == "0":  # program
                    self.program_container.websocket_event(event)
                elif event.action == "6":  # variable change
                    self.variable_container.websocket_event(event)

            elif event.control == "_5":  # system status
                self.process_system_status(event)

        except Exception as ex:
            logger.error("websocket handler Error {}".format(ex))
            traceback.print_exc()

    def process_controller_event(self, property_, value):
        controller = self.controller_container.get("controller")
        controller.set_property(property_, value)

    def process_heartbeat(self, event):
        self.last_heartbeat = datetime.now()  # .strftime("%m/%d/%Y, %H:%M:%S")
        self.process_controller_event("heartbeat", self.last_heartbeat)
        self.heartbeat_interval = int(event.action)

    def process_system_status(self, event):
        if event.action == "0":  # idle
            self.process_controller_event("state", "idle")
        elif event.action == "1":  # busy
            self.process_controller_event("state", "busy")

    def watch_dog_check(self):
        if (
            self.last_heartbeat + timedelta(seconds=self.heartbeat_interval)
            < datetime.now()
        ):
            logger.warning("Watchdog timer triggered.")

    def close(self):
        self.event_loop.stop()
        self._ws_thread.join()
        self.session.close()


#! /usr/bin/env python


from .device_base import Device_Base


class Device_Speed_Controller(Device_Base):
    def __init__(
        self, container, name, address, speeds=["off", "low", "medium", "high"]
    ):
        Device_Base.__init__(self, container, "speed_controller", name, address)

        self.speeds = speeds

        self.add_property("speed", "off")

    def set_speed(self, speed):
        pass




#! /usr/bin/env python

from ..item_container import Item_Container
from .controller_base import Controller_Base


import logging

logger = logging.getLogger(__name__)


class Controller_Container(Item_Container):
    def __init__(self, controller):
        Item_Container.__init__(self, controller, "Controller")

    def start(self):

        controller = Controller_Base(self, "controller")
        self.add(controller, controller.id)

        self.items_retrieved = True

        return True

    def controller_event(self, property, value):
        controller = self.get("controller")
        controller.set_property(property, value)

    def get_controller_time(self):
        controller = self.get("controller")
        return controller.get_controller_time()



#! /usr/bin/env python


from .device_base import Device_Base


class Device_Thermostat(Device_Base):
    def __init__(self, container, name, address):
        Device_Base.__init__(self, container, "thermostat", name, address)

        self.add_property("temperature", 0)
        self.add_property("humidity", 0)
        self.add_property("coolsetpoint", 0)
        self.add_property("heatsetpoint", 0)
        self.add_property("systemmode", 0)
        self.add_property("systemstatus", "Idle")
        self.add_property("fanmode", 0)

    def set_mode(self, mode):
        pass

    def set_status(self, status):
        pass

    def set_heatsetpoint(self, setpoint):
        pass

    def set_coolsetpoint(self, setpoint):
        pass

    def set_fanmode(self, mode):
        pass



#! /usr/bin/env python

import xml.etree.ElementTree as ET

import traceback

# process a node and assemble the info to build a device


""" zwave node....
<nodeInfo>
<node flag="0">
<address>ZW002_143</address>
<name>ZW 002 Energy Meter</name>
<family>4</family>
<type>4.33.1.0</type>
<enabled>true</enabled>
<deviceClass>0</deviceClass>
<wattage>0</wattage>
<dcPeriod>0</dcPeriod>
<startDelay>0</startDelay>
<endDelay>0</endDelay>
<pnode>ZW002_1</pnode>
<sgid>143</sgid>
<devtype>
<gen>4.33.1</gen>
<mfg>134.2.9</mfg>
<cat>143</cat>
</devtype>
<ELK_ID>I16</ELK_ID>
<property id="ST" value="91289" formatted="912.89 Watts" uom="73" prec="2"/>
</node>
<properties>
<property id="ST" value="91289" formatted="912.89 Watts" uom="73" prec="2"/>
<property id="TPW" value="31765" formatted="317.65 kWh" uom="33" prec="2"/>
</properties>
</nodeInfo>
"""


class Device_Info(object):
    def __init__(self, node):

        self.node = node

        self.valid = False

        self.properties = {}

        try:
            self.family = None
            family_ = node.find("family")
            if family_ is not None:
                self.family = family_.text
            else:
                self.family = "1"

            type_ = node.find("type")
            self.type = type_.text

            types = type_.text.split(".")
            self.category = types[0]
            self.sub_category = types[1]
            self.version = types[2]
            self.reserved = types[3]

            address = node.find("address")
            self.address = address.text
            self.address_parts = address.text.split(" ")

            name = node.find("name")
            self.name = name.text

            flag = node.get("flag")
            self.flag = int(flag)
            self.node_def_id = node.get("nodeDefId")

            # container_node = node.find('container')
            # self.container_node_address = container_node.text

            for node_property in node.iter("property"):
                id = node_property.attrib["id"]
                property_ = {}
                for key, value in node_property.items():
                    if key != "id":
                        property_[key] = value
                self.properties[id] = property_

            self.devtype_cat = None
            devtype_node = node.find("devtype")
            if devtype_node is not None:
                devtype_cat_ = devtype_node.find("cat")
                if devtype_cat_ is not None:
                    self.devtype_cat = devtype_cat_.text

            self.valid = True

            # print (self)

        except Exception:
            traceback.print_exc()

    def get_property(self, node, key):
        if node in self.properties:
            if key in self.properties[node]:
                return self.properties[node][key]
        return None

    def __repr__(self):
        return "Device: Name {} Address {}, Family {}, Type {}, DevCat {}, Def ID {}".format(
            self.name,
            self.address,
            self.family,
            self.type,
            self.devtype_cat,
            self.node_def_id,
        )




#! /usr/bin/env python


from ..common.device_power import Device_Power
from .device_zwave_base import Device_ZWave_Base


class Device_ZWave_Power(Device_Power, Device_ZWave_Base):
    def __init__(self, container, device_info):
        Device_Power.__init__(self, container, device_info.name, device_info.address)
        Device_ZWave_Base.__init__(self, device_info)

        value = device_info.get_property("ST", "value")
        if value:
            try:
                self.set_property("power", value)
            except:
                pass

        # could capture other properties -> uofm, prec, format

    def process_websocket_event(self, event):
        if event.control == "ST":
            self.set_property("power", event.action)




#! /usr/bin/env python


from .device_base import Device_Base


class Device_Lock(Device_Base):
    def __init__(self, container, name, address):
        Device_Base.__init__(self, container, "lock", name, address)

        self.add_property("batterylevel", 0)
        self.add_property("state", 0)
        self.add_property("error", 0)
        self.add_property("usernumber", 0)

    def lock(self):
        pass

    def unlock(self):
        pass



#! /usr/bin/env python

""" 

Base Scene

Common to all elements returned from rest/nodes/scenes

TBD - may need to change - scene_base and scene_inteon based on insteon only setup

"""

from ..item_base import Item_Base


class Scene_Base(Item_Base):
    def __init__(self, container, scene_info):
        Item_Base.__init__(self, container, scene_info.name)

        self.address = scene_info.address
        self.flag = scene_info.flag
        self.id = scene_info.id
        self.family = scene_info.family

        self.primary_node = scene_info.primary_node

        self.device_group = scene_info.device_group

        self.controllers = scene_info.controllers
        self.responders = scene_info.responders

    def __str__(self):
        return "Scene: {} ; address {} ; flag {}; id {}".format(
            self.name, self.address, self.flag, self.id
        )

    def device_event(
        self, device
    ):  # device event, process and see if we are interested
        pass  # subclasses to provide

    def get_identifier(self):
        return self.address




#! /usr/bin/env python

# process an event from ISY web socket client


class Websocket_Event(object):
    def __init__(self, event):

        self.event = event

        self.valid = False

        try:
            node = event.find("node")
            control = event.find("control")
            action = event.find("action")
            event_info = event.find("eventInfo")

            self.address = node.text
            self.control = control.text
            self.action = action.text
            self.event_info = event_info.text

            self.event_info_node = event_info  # store as consumers may need to use it

            self.valid = True

        except:
            pass

    def __repr__(self):
        return "Event: Address {}, Control {}, Action {}, Event Info {}".format(
            self.address, self.control, self.action, self.event_info
        )




#! /usr/bin/env python
import xml.etree.ElementTree as ET

from ..item_container import Item_Container
from .variable_info import Variable_Info
from .variable_name import Variable_Name
from .variable_integer import Variable_Integer
from .variable_state import Variable_State

import logging

logger = logging.getLogger(__name__)

variable_classes = {
    "1": Variable_Integer,
    "2": Variable_State,
}


class Variable_Container(Item_Container):
    def __init__(self, controller):
        Item_Container.__init__(self, controller, "Variable")

    def get_list(self, request_string):
        success, response = self.send_request(request_string, 2)

        if not success:
            raise Exception("no response for request: ", request_string)
        else:
            return ET.fromstring(response)

    def get_and_process(self, var_type):
        try:
            vars_list = self.get_list("vars/get/" + var_type)
            defs_list = self.get_list("vars/definitions/" + var_type)
            self.process_variable_nodes(vars_list, defs_list)
            return True
       
        except Exception as error:
            logger.error("Error processing variabls of  type {}: {}".format(var_type, error))
            return False

    def start(self):
        success = True
        if self.get_and_process("1") is False:  # integer variables
            success = False
        if self.get_and_process("2") is False: # state variables
            success = False
        self.items_retrieved = success
        return success

    def process_variable_nodes(self, list_root, name_root):
        for node in list_root.iter("var"):
            variable_info = Variable_Info(node)

            if variable_info.valid:  # make sure we have the info we need
                variable_class = variable_classes[variable_info.type]
                variable = variable_class(self, variable_info)

                # find name
                for node in name_root.iter("e"):
                    variable_name = Variable_Name(node)

                    if variable_name.valid and variable_name.id == variable.id:
                        variable.name = variable_name.name

                self.add(variable, variable.get_index())

    def websocket_event(self, event):
        var = event.event_info_node.find("var")
        variable_id = var.attrib["id"]
        variable_type = var.attrib["type"]

        index = variable_type + ":" + variable_id

        variable = self.get(index)
        if variable is not None:
            variable.process_websocket_event(event)
        else:
            logger.warning(
                "Unable able to find variable type {} id {}".format(
                    variable_id, variable_type
                )
            )




#! /usr/bin/env python

""" 

Base Device

Common to all ISY items ie programs, variables, devices, scenes

status property -> ready, alert, lost

provides a common mechanism to subscribe to item property changes

"""


class Item_Base(object):
    def __init__(self, container, name):
        self.container = container

        self.name = name

        self.properties = {
            "status": "init"
        }  # list of properties key = property name, value = property value

        self.property_event_handlers = []

    def process_websocket_event(self, event):
        pass  # classes to override

    def add_property(self, property_, value=None):
        self.properties[property_] = value

    def set_property(
        self, property_, value, always_publish=False
    ):  # propagates up to the container and to handlers
        if self.properties[property_] != value or always_publish:
            self.properties[property_] = value
            self.container.property_change(self, property_, value)

            for handler in self.property_event_handlers:
                handler(property_, value)

    def get_property(self, property_):
        return self.properties[property_]

    def send_request(self, path, timeout=None):
        return self.container.send_request(path, timeout)

    def add_property_event_handler(self, handler):
        self.property_event_handlers.append(handler)

    def get_identifier(self):
        pass  # subclasses to provide




#!/usr/local/bin/python
import socket
import struct
from threading import Thread, Timer

TIMEOUT = 1.5


class Discover(object):
    def __init__(self):
        pass

    def start(self):
        ISYs = []
        MCAST_GRP = "239.255.255.250"
        MCAST_PORT = 1900
        msg = 'M-SEARCH * HTTP/1.1\r\nHOST:239.255.255.250:1900\r\nMAN:"ssdp.discover"\r\nMX:1\r\nST:urn:udi-com:device:X_Insteon_Lighting_Device:1\r\n\r\n'.encode()

        self.discovery_done = False

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 5)
        mreq = struct.pack("4sl", socket.inet_aton(MCAST_GRP), socket.INADDR_ANY)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        sock.settimeout(0.1)

        def multicast_timeout():
            self.discovery_done = True

        timer = Timer(TIMEOUT, multicast_timeout)
        timer.start()
        sock.sendto(msg, (MCAST_GRP, MCAST_PORT))

        while self.discovery_done != True:
            try:
                data = sock.recv(10240).decode()

                if "HTTP/1.1 200 OK" in data:
                    index = data.index("LOCATION:")
                    index2 = data.index("/desc")
                    location = data[index + 16 : index2]
                    index = data.index("USN:")
                    index2 = data.index("::urn")
                    uuid = data[index + 4 : index2]
                    ISY = "%s [%s]" % (location, uuid)
                    # print('isy found',ISY)
                    ISYs.append(location)
            except Exception as e:
                pass

        sock.close()
        return ISYs


if __name__ == "__main__":
    d = Discover()
    isys = d.start()

    print(isys)

    for x in isys:
        print(x)



#! /usr/bin/env python


from ..common.device_dimmer import Device_Dimmer
from .device_insteon_base import Device_Insteon_Base

paddle_events = {"DON", "DOF", "DIM", "BRT", "DFON", "DFOF"}


class Device_Insteon_Dimmer(Device_Dimmer, Device_Insteon_Base):
    def __init__(self, container, device_info):
        Device_Dimmer.__init__(self, container, device_info.name, device_info.address)
        Device_Insteon_Base.__init__(self, device_info)
        # print(device_info)

        self.add_property("paddle_action")

        value = device_info.get_property("ST", "value")
        if value:
            try:
                self.properties["level"] = int(int(value) / 255 * 100)
            except:
                pass

    def process_websocket_event(self, event):
        if event.control == "ST":
            self.set_property("level", int(int(event.action) / 255 * 100))
            # print ('device {}. changed status to {}'.format(self.name,event.action))

        elif event.control in paddle_events:  # need to add other events
            self.set_property("paddle_action", event.control, True)
            # print ('device {}. changed local control {}'.format(self.name,event.action))

    def set_level(self, level):
        path = "nodes/" + self.address + "/cmd/DON/" + str(int(level / 100 * 255))
        return self.send_request(path)

    def fast_on(self):
        path = "nodes/" + self.address + "/cmd/DFON"
        return self.send_request(path)

    def fast_off(self):
        path = "nodes/" + self.address + "/cmd/DFOF"
        return self.send_request(path)

    def brighten(self):
        path = "nodes/" + self.address + "/cmd/BRT"
        return self.send_request(path)

    def dim(self):
        path = "nodes/" + self.address + "/cmd/DIM"
        return self.send_request(path)



#! /usr/bin/env python


# process a node and assemble the info to build a program


class Program_Info(object):
    def __init__(self, node):

        self.node = node

        self.valid = False

        try:
            if node.attrib["folder"] == "true":  # program folder, not interested
                return

            #if node.attrib["enabled"] == "false":  # not enabled
            #    return

            self.id = node.attrib["id"]
            self.name = node.find("name").text
            self.last_run_time = node.find("lastRunTime").text
            self.last_finish_time = node.find("lastFinishTime").text

            self.valid = True
        except:
            pass

    def __repr__(self):
        return "Program: Name {} ID {}".format(self.name, self.id)




#!/usr/local/bin/python

import asyncio
import functools
import logging
import xml.etree.ElementTree as ET

import aiohttp

from .websocket_event import Websocket_Event

logger = logging.getLogger(__name__)

ws_headers = {
    "Sec-WebSocket-Protocol": "ISYSUB",
    "Origin": "com.universal-devices.websockets.isy",
}

http_headers = {
    "Connection": "keep-alive",
}


class Async_Session(object):
    def __init__(
        self, controller, address, port, username, password, https=False, loop=False
    ):
        self.controller = controller

        self._address = address
        self._port = port
        self._https = https  # not implemented

        self.loop = loop

        # time outs
        self._heart_beat = 30
        self._request_timeout = 30
        self._sleep_time = 10

        self._websocket_connected = False
        self._websocket_url = "ws://" + self._address + "/rest/subscribe"

        self._http_connected = False
        self._rest_url = "http://" + self._address + "/rest/"

        self.session = None
        self.auth = aiohttp.BasicAuth(username, password)
        self.timeout = aiohttp.ClientTimeout(
            total=60, connect=60, sock_connect=60, sock_read=60
        )
        self.loop.run_until_complete(self.create_new_session())

        self.keep_listening = True

    @property
    def websocket_connected(self):
        return self._websocket_connected

    @websocket_connected.setter
    def websocket_connected(self, connected):
        if self._websocket_connected != connected:
            self._websocket_connected = connected
            logger.info("Websocket Connected {}".format(connected))
            if self.controller:
                wrapped = functools.partial(
                    self.controller.websocket_connected, connected
                )
                self.loop.call_soon(wrapped)

    @property
    def http_connected(self):
        return self._http_connected

    @http_connected.setter
    def http_connected(self, connected):
        if self._http_connected != connected:
            self._http_connected = connected
            logger.info("HTTP Connected {}".format(connected))
            if self.controller:
                wrapped = functools.partial(self.controller.http_connected, connected)
                self.loop.call_soon(wrapped)

    async def create_new_session(self):
        if self.session is not None and self.session.closed is False:
            await self.session.close()

        self.session = aiohttp.ClientSession(
            auth=self.auth,
            raise_for_status=False,
            headers=http_headers,
            timeout=self.timeout,
        )

    async def request_async(self, path, timeout):
        timeout = 60#timeout or self._request_timeout

        if self.session is None:
            await self.create_new_session()

        tries = 0
        while tries < 5:
            try:
                tries = tries +1
                logger.info("HTTP Get to {}".format(self._rest_url + path))

                async with self.session.get(
                    self._rest_url + path, chunked=True, timeout=timeout
                ) as response:
                    if response.status == 200:
                        self.http_connected = True
                        body = await response.text()
                        logger.info("HTTP Get response {}".format(body))
                        # print(body)
                        return True, body
                
                logger.warning ("Retrying HTTP Get Failed {} Try # {}".format(response,tries))
                await asyncio.sleep(.5)

            except Exception as ex:
                logger.warning("HTTP Get Error {}".format(ex))
                self.http_connected = False
                await asyncio.sleep(.5)

        logger.error("HTTP request failed after {} tries. Status from last try error {}".format(tries,response.status))
        return False, None

    def request(self, path, timeout):  # sync
        timeout = timeout or self._request_timeout

        try:
            future = asyncio.run_coroutine_threadsafe(
                self.request_async(path, timeout), self.loop
            )
            return future.result(10)

        except Exception as ex:
            logger.error("HTTP Get Error {}".format(ex))
            return False, None

    def start_websocket(self):
        self.websocket_task = asyncio.run_coroutine_threadsafe(self.listen_forever(),self.loop)

    async def listen_forever(self):
        logger.info("WebSocket Listen Forever Start")

        await asyncio.sleep(1)

        while self.keep_listening:

            try:
                logger.info("Connecting to WebSocket")

                try:
                    async with self.session.ws_connect(
                        self._websocket_url,
                        headers=ws_headers,
                        auth=self.auth,
                        heartbeat=self._heart_beat,
                        receive_timeout=self._request_timeout,
                    ) as ws:
                        logger.info("Websocket waiting for messages")

                        async for msg in ws:
                            self.websocket_connected = True
                            # print('Message received from server:', msg)
                            logger.debug("Websocket Message: {}".format(msg))
                            if msg.type == aiohttp.WSMsgType.TEXT:
                                try:
                                    event_node = ET.fromstring(msg.data)

                                    if event_node.tag == "Event":
                                        event = Websocket_Event(event_node)

                                        if event.valid:
                                            logger.debug("Websocket Event {}".format(event))
                                            if self.controller:
                                                wrapped = functools.partial(
                                                    self.controller.websocket_event, event
                                                )
                                                self.loop.call_soon(wrapped)

                                except Exception as ex:
                                    logger.error("Websocket Message Error {}".format(ex))

                            elif msg.type == aiohttp.WSMsgType.BINARY:
                                logger.warning("Websocket Binary: {}".format(msg.data))

                            elif msg.type == aiohttp.WSMsgType.PING:
                                logger.warning("Ping received")
                                ws.pong()

                            elif msg.type == aiohttp.WSMsgType.PONG:
                                logger.warning("Pong received")

                            elif msg.type == aiohttp.WSMsgType.CLOSE:
                                logger.warning("Close received")
                                await ws.close()

                            elif msg.type == aiohttp.WSMsgType.ERROR:
                                logger.error(
                                    "Error during receive {}".format(ws.exception())
                                )

                            elif msg.type == aiohttp.WSMsgType.CLOSED:
                                logger.warning("Close {}".format(ws.exception()))
                                await ws.close()

                    self.websocket_connected = False

                    logger.warning("Websocket Aysnc loop completed.")
                    await asyncio.sleep(self._sleep_time)

                except Exception as ex:
                    self.websocket_connected = False
                    logger.error("Websocket Error {}".format(ex))
                    await asyncio.sleep(self._sleep_time)
                    continue

                if self.keep_listening:
                    await asyncio.sleep(self._sleep_time)
                    await self.create_new_session()

            except Exception as ex:
                logger.error("Inner listen forever error {}".format(ex))

    def close(self):
        self.websocket_connected = False
        self.http_connected = False

        self.keep_listening = False

        self.websocket_task.cancel()

        if self.session is not None and self.session.closed is False:
            self.loop.run_until_complete(self.session.close())

if __name__ == "__main__":
    event_loop = asyncio.get_event_loop()
    wsc = Async_Session(
        False, "192.168.1.51", 80, "admin", "admin", False, loop=event_loop
    )

    try:
        wsc.start_websocket()
        event_loop.run_forever()

    except KeyboardInterrupt:
        print("KeyboardInterrupt has been caught.")
        wsc.close()

    finally:
        event_loop.close()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name='ISY994v5',
    version='0.8.7',
    description='ISY99 Controller Rest and Websocket client v5 firmware',
    author='Michael Cumming',
    author_email='mike@4831.com',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/mjcumming/ISY994v5',
    keywords=['INSTEON', 'ISY994', 'ISY', 'Universal Devices'],
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'aiohttp',
    ]
)



#! /usr/bin/env python


from ..common.device_switch import Device_Switch
from .device_insteon_base import Device_Insteon_Base


class Device_Insteon_Siren(Device_Switch, Device_Insteon_Base):
    def __init__(self, container, device_info):
        Device_Switch.__init__(self, container, device_info.name, device_info.address)
        Device_Insteon_Base.__init__(self, device_info)

        self.add_property("duration")
        self.add_property("mode")

        value = device_info.get_property("ST", "value")
        if value:
            try:
                if int(value) > 0:
                    self.set_property("onoff", "on")
                else:
                    self.set_property("onoff", "off")
            except:
                pass

    def process_websocket_event(self, event):
        if event.control == "ST":
            if int(event.action) > 0:
                self.set_property("onoff", "on")
            else:
                self.set_property("onoff", "off")
        elif event.control == "DUR":
            self.set_property("duration",int(event.action))
        elif event.control == "MODE":
            self.set_property("mode",int(event.action))

    def turn_on(self):
        path = "nodes/" + self.address + "/cmd/DON"
        return self.send_request(path)

    def turn_off(self):
        path = "nodes/" + self.address + "/cmd/DOF"
        return self.send_request(path)



#! /usr/bin/env python


from ..common.device_thermostat import Device_Thermostat
from .device_insteon_base import Device_Insteon_Base

MODES = ["Off", "Heat", "Cool", "Auto", "Program Auto","Program Auto"]
STATUS = ["Idle", "Heating", "Cooling"]

FAN_MODES = {
    0: "Off",
    7: "On",
    8: "Auto",
}


class Device_Insteon_TempLinc(Device_Thermostat, Device_Insteon_Base):
    def __init__(self, container, device_info):
        Device_Thermostat.__init__(
            self, container, device_info.name, device_info.address
        )
        Device_Insteon_Base.__init__(self, device_info)

        coolsetpoint_address_parts = device_info.address_parts.copy()
        coolsetpoint_address_parts[3] = "2"
        self.coolsetpoint_address = " ".join(coolsetpoint_address_parts)

        heatsetpoint_address_parts = device_info.address_parts.copy()
        heatsetpoint_address_parts[3] = "2"
        self.heatsetpoint_address = " ".join(heatsetpoint_address_parts)

        # print ('device properties {}'.format(device_info.properties))
        value = device_info.get_property("ST", "value")
        if value:
            try:
                self.set_property("temperature", round(float(value) / 2, 0))
            except:
                self.set_property("temperature", 0)

        value = device_info.get_property("CLIMD", "value")
        if value:
            try:
                self.set_property("systemmode", MODES[int(value)])
            except:
                self.set_property("systemmode", 0)

        value = device_info.get_property("CLISPH", "value")
        if value:
            try:
                self.set_property("heatsetpoint", round(float(value) / 2, 0))
            except:
                self.set_property("heatsetpoint", 0)

        value = device_info.get_property("CLISPC", "value")
        if value:
            try:
                self.set_property("coolsetpoint", round(float(value) / 2, 0))
            except:
                self.set_property("coolsetpoint", 0)

        value = device_info.get_property("CLIHUM", "value")
        if value:
            try:
                self.set_property("humidity", round(float(value), 0))
            except:
                self.set_property("humidity", 0)

        value = device_info.get_property("CLIFS", "value")
        if value:
            try:
                self.set_property("fanmode", FAN_MODES[int(value)])
            except:
                self.set_property("fanmode", 0)
        value = device_info.get_property("CLIHCS", "value")
        if value:
            try:
                self.set_property("systemstatus", STATUS[int(value)])
            except:
                self.set_property("systemstatus", 0)

    def process_websocket_event(self, event):
        if event.control == "ST":
            self.set_property("temperature", round(float(event.action) / 2, 0))
        elif event.control == "CLISPH":
            self.set_property("heatsetpoint", round(float(event.action) / 2, 0))
        elif event.control == "CLISPC":
            self.set_property("coolsetpoint", round(float(event.action) / 2, 0))
        elif event.control == "CLIHUM":
            self.set_property("humidity", round(float(event.action), 0))
        elif event.control == "CLIMD":
            self.set_property("systemmode", MODES[int(event.action)])
        elif event.control == "CLIFS":
            self.set_property("fanmode", FAN_MODES[int(event.action)])
        elif event.control == "CLIHCS":
            self.set_property("systemstatus", STATUS[int(event.action)])

    def set_mode(self, mode):
        path = "nodes/" + self.address + "/cmd/DON/" + str(mode)
        return self.send_request(path)

    def set_heatsetpoint(self, setpoint):
        path = "nodes/" + self.heatsetpoint_address + "/cmd/DON/" + str(setpoint * 2)
        return self.send_request(path)

    def set_coolsetpoint(self, setpoint):
        path = "nodes/" + self.coolsetpoint_address + "/cmd/DON/" + str(setpoint * 2)
        return self.send_request(path)



#! /usr/bin/env python

"""

container for a collection of items

container_type = device, variable, program, scene, controller

publishes all item events back to the controller object

"""

import copy
import logging

logger = logging.getLogger(__name__)

# containter events = {'add','remove','property'}


class Item_Container(object):
    def __init__(self, controller, container_type):
        self.controller = controller
        self.container_type = container_type
        self.items_retrieved = False

        self.list = {}  # indexed by device(node) address

        self.started = False

    def start(self):
        pass

    def send_request(self, path, timeout=None):
        return self.controller.send_request(path, timeout)

    def websocket_event(self, event):
        pass

    def add(self, item, key):
        self.list[key] = item
        self.event(item, "add")

    def remove(self, key):
        item = self.list[key]
        del self.list[key]
        self.event(item, "remove")

    def get(self, key):
        if key in self.list:
            return self.list[key]
        else:
            return None

    def get_list_copy(self):
        return copy.copy(self.list)

    def property_change(
        self, item, property_, value
    ):  # called by the item to publish a change
        self.event(item, "property", property_, value)

    def event(self, item, event, *args):  # publish event to controller
        # print ('item event',self.container_type,item.name,event)
        self.controller.container_event(self, item, event, args)
